"""
Plugin for Beancount to share expenses.

Credits: Based on Martin Blais <blais@furius.ca> personal snippet: https://github.com/beancount/beancount/issues/474.
"""
__author__ = "Akuukis"

from datetime import date, datetime
from typing import NamedTuple, Set, List, Union, Tuple
from collections import namedtuple
import re

from beancount.core.inventory import Inventory
from beancount.core.number import D, Decimal, ONE
from beancount.core.amount import Amount
from beancount.core.data import Account, Entries, Posting, Open, Transaction, new_metadata

from beancount_share.common import read_config, normalize_marked_txn, marked_postings, sum_income, sum_expenses
import beancount_share.metaset as metaset

__plugins__ = ['share']

# pylint: disable=raising-non-exception

Config = NamedTuple('Config', [
    ('mark_name', str),
    ('meta_name', Union[str, None]),
    ('account_debtors', str),
    ('account_creditors', str),
    ('quantize', Decimal),
    ('open_date', Union[date, None]),
])

PluginShareParseError = namedtuple('LoadError', 'source message entry')


def share(entries: Entries, unused_options_map, config_string="{}") -> Tuple[Entries, List[Exception]]:
    new_accounts: Set[Account] = set()
    new_entries: Entries = []
    errors: List[PluginShareParseError] = []

    # 1. Parse config
    config_obj = read_config(config_string)
    raw_open_date = config_obj.pop('open_date', '1970-01-01')
    config = Config(
        config_obj.pop('mark_name'         , 'share'),
        config_obj.pop('meta_name'         , 'shared'),
        config_obj.pop('account_debtors'   , 'Assets:Debtors'),
        config_obj.pop('account_creditors' , 'Liabilities:Creditors'),
        D(str(config_obj.pop('quantize'    , 0.01))),
        None if raw_open_date is None else date.fromisoformat(raw_open_date),
    )

    for entry in entries:
        if not isinstance(entry, Transaction):
            new_entries.append(entry)
            continue

        # 2. Normalize marks.
        tx = normalize_marked_txn(entry, config.mark_name)
        marked_postings_result = list(marked_postings(tx, config.mark_name, ("Income", "Expenses")))

        # 3. Determine whenever this is debtor or creditor transactions.
        account_prefix: str
        total_income = sum_income(tx)
        total_expenses = sum_expenses(tx)
        total_value: Amount

        # 3.1. If transaction is marked but has no effect on any postings, error.
        if(total_expenses.is_empty() and total_income.is_empty() and metaset.get(tx.meta, config.mark_name)):
            errors.append(PluginShareParseError(
                new_metadata(entry.meta['filename'], entry.meta['lineno']),
                "Plugin \"share\" doesn't work on transactions that has nor income and expense.",
                entry,
            ))
            new_entries.append(entry)
            continue

        # 3.2. If tx nor postings are not marked, bail early.
        if(len([elem for elem,_,_,_ in marked_postings_result if elem != None]) == 0):
            new_entries.append(entry)
            continue

        if(not total_expenses.is_empty() and total_income.is_empty()):
            account_prefix = config.account_debtors + ':'
            total_value = total_expenses.get_currency_units(tx.postings[0].units.currency)
        elif(total_expenses.is_empty() and not total_income.is_empty()):
            account_prefix = config.account_creditors + ':'
            total_value = total_income.get_currency_units(tx.postings[0].units.currency)
        else:
            errors.append(PluginShareParseError(
                new_metadata(entry.meta['filename'], entry.meta['lineno']),
                "Plugin \"share\" doesn't work on transactions that has both income and expense: please split it up into two transactions instead.",
                entry,
            ))
            new_entries.append(entry)
            continue

        # 4. Per posting, split it up based on marks.
        new_postings = []
        for marks, posting, orig, tx_clean in marked_postings_result:

            # 4.1. or skip if not marked.
            if(marks == None):
                new_postings.append(posting)
                continue

            if(not (posting.account.split(':')[0] in ("Income", "Expenses"))):
                errors.append(PluginShareParseError(
                    new_metadata(posting.meta['filename'], posting.meta['lineno']),
                    'Mark "share" doesn\'t make sense on a "{}" type posting.'.format(posting.account.split(':')[0]),
                    entry,
                ))
                new_postings.append(orig)
                continue

            # 5. Per mark, create a new posting.
            todo_absolute: List[Tuple[Amount, str]] = list()
            todo_percent: List[Tuple[float, str]] = list()
            todo_absent: List[str] = list()
            for mark in marks:
                parts = mark.split('-')
                account: str

                # 5.1. Apply defaults.
                try:
                    account = parts[0] if ':' in parts[0] else account_prefix + parts[0]
                    new_accounts.add(account)

                except IndexError:
                    errors.append(PluginShareParseError(
                        new_metadata(posting.meta['filename'], posting.meta['lineno']),
                        "Mark \"{}\" must contain account name.".format(mark),
                        entry,
                    ))
                    new_postings.append(orig)
                    continue

                if(len(parts) > 1):
                    if('%' in parts[1] or 'p' in parts[1]):
                        try:
                            todo_percent.append((float(parts[1].split('%')[0].split('p')[0])/100, account))
                        except IndexError:
                            errors.append(PluginShareParseError(
                                new_metadata(posting.meta['filename'], posting.meta['lineno']),
                                "Something wrong with percent fraction \"{}\", please use a dot, e.g. \"33.3%\".".format(parts[1]),
                                entry,
                            ))
                            new_postings.append(orig)
                            continue
                    else:
                        try:
                            todo_absolute.append((Amount(D(parts[1]).quantize(config.quantize), posting.units.currency), account))
                        except IndexError:
                            errors.append(PluginShareParseError(
                                new_metadata(posting.meta['filename'], posting.meta['lineno']),
                                "Something wrong with absolute fraction \"{}\", please use a dot, e.g. \"33.3\".".format(parts[1]),
                                entry,
                            ))
                            new_postings.append(orig)
                            continue
                else:
                    todo_absent.append(account)

            total_shared_absolute = sum([amount.number for amount,_ in todo_absolute], D(0).quantize(config.quantize))
            total_shared_relative = sum([percent for percent,_ in todo_percent])

            if(total_shared_absolute > abs(total_value.number)):
                errors.append(PluginShareParseError(
                    new_metadata(posting.meta['filename'], posting.meta['lineno']),
                    "The posting can't share more than it's absolute value",
                    entry,
                ))
                new_postings.append(orig)
                continue

            if(total_shared_relative > 1):
                errors.append(PluginShareParseError(
                    new_metadata(posting.meta['filename'], posting.meta['lineno']),
                    "The posting can't share more percent than 100%.",
                    entry,
                ))
                new_postings.append(orig)
                continue

            if(total_shared_absolute == abs(total_value.number) and total_shared_relative > 0):
                errors.append(PluginShareParseError(
                    new_metadata(posting.meta['filename'], posting.meta['lineno']),
                    "It doesn't make sense to split a remaining amount of zero.",
                    entry,
                ))
                new_postings.append(orig)
                continue

            if(total_shared_relative == 1 and len(todo_absent) > 0):
                errors.append(PluginShareParseError(
                    new_metadata(posting.meta['filename'], posting.meta['lineno']),
                    "It doesn't make sense to further auto-split when amount is already split for full 100%.",
                    entry,
                ))
                new_postings.append(orig)
                continue


            new_postings_inner = []
            # 5.2. Handle absolute amounts first: mutate original posting's amount & create new postings.
            for amount, account in todo_absolute:
                posting = posting._replace(
                    units=posting.units._replace(number=(posting.units.number - amount.number).quantize(config.quantize)),
                )
                if config.meta_name is not None:
                    posting = posting._replace(meta=metaset.add(posting.meta, config.meta_name, account + " " + amount.to_string()))
                new_postings_inner.append(Posting(
                    account,
                    units=posting.units._replace(number=(amount.number).quantize(config.quantize)),
                    cost=posting.cost,
                    price=None,
                    flag=None,
                    meta={} if config.meta_name is None else {config.meta_name: posting.account + " " + amount.to_string()}
                ))

            # 5.3. Handle relative amounts second: create new postings.
            remainder = posting.units
            total = D(0)
            for percent, account in todo_percent:
                units = posting.units._replace(number=(D(float(remainder.number) * percent)).quantize(config.quantize))
                total = total + units.number
                new_postings_inner.append(Posting(
                    account,
                    units=units,
                    cost=posting.cost,
                    price=None,
                    flag=None,
                    meta={} if config.meta_name is None else {config.meta_name: posting.account + " " + str(int(percent * 100))+'% (' + units.to_string() + ')'}
                ))
                if config.meta_name is not None:
                    posting = posting._replace(meta=metaset.add(posting.meta, config.meta_name, account + " " + str(int(percent * 100))+'% (' + units.to_string() + ')'))

            # 5.4. Handle absent amounts third: create new postings.
            total_percent = sum(i for i, _ in todo_percent)
            percent = (1 - total_percent) / (1 + len(todo_absent))
            for account in todo_absent:
                units = posting.units._replace(number=(D(float(remainder.number) * percent)).quantize(config.quantize))
                total = total + units.number
                new_postings_inner.append(Posting(
                    account,
                    units=units,
                    cost=posting.cost,
                    price=None,
                    flag=None,
                    meta={} if config.meta_name is None else {config.meta_name: posting.account + " (" + str(int(percent * 100))+'%, ' + units.to_string() + ')'}
                ))
                if config.meta_name is not None:
                    posting = posting._replace(meta=metaset.add(posting.meta, config.meta_name, account + " (" + str(int(percent * 100))+'%, ' + units.to_string() + ')'))

            # 5.5. Handle original posting last (mutate!).
            posting = posting._replace(
                units=posting.units._replace(number=(remainder.number - total).quantize(config.quantize))
            )

            # if(posting.units.number > D(0)):
            new_postings.append(posting)

            new_postings.extend(new_postings_inner)

        for account in new_accounts:
            new_postings = group_postings(new_postings, account, config.meta_name)

        new_entries.append(tx_clean._replace(
            postings=new_postings,
        ))

    if(config.open_date != None):
        for account in sorted(new_accounts):
            new_meta = new_metadata(entries[0].meta['filename'], 0)
            open_entry = Open(new_meta, config.open_date, account, None, None)
            new_entries.append(open_entry)

    return new_entries, errors


def group_postings(postings: List[Posting], account: Account, meta_name: Union[str, None]) -> List[Posting]:
    grouped_postings = []
    share_postings = []
    share_balance = Inventory()
    meta = dict()
    for posting in postings:
        if posting.account == account:
            share_postings.append(posting)
            share_balance.add_position(posting)
            if(len(meta) == 0):
                meta = posting.meta
            elif(meta_name is not None):
                for mark in metaset.get(posting.meta, meta_name):
                    share_postings[0] = share_postings[0]._replace(meta=metaset.add(share_postings[0].meta, meta_name, mark))
        else:
            grouped_postings.append(posting)

    if share_postings:
        for pos in share_balance:
            grouped_postings.append(Posting(account, pos.units, pos.cost, None, None, share_postings[0].meta))

    return grouped_postings



from typing import List, Set, Union, Tuple

from beancount.core.data import Transaction, Posting
from beancount.core.inventory import Inventory

import beancount_share.metaset as metaset

def read_config(config_string):
    """
    Args:
      config_string: A configuration string in JSON format given in source file.
    Returns:
      A dict of the configuration string.
    """
    if len(config_string) == 0:
        config_obj = {}
    else:
        config_obj = eval(config_string, {}, {})

    if not isinstance(config_obj, dict):
        raise RuntimeError("Invalid plugin configuration: should be a single dict.")
    return config_obj

MARK_SEPERATOR = '-'
def normalize_marked_txn(tx: Transaction, mark_name: str):
    """
    If a transaction is marked, hoist marked tags into transation meta(s).

    Args:
        txs [Transaction]: transaction instances.
        mark_name [str]: the mark.
    Return:
        Transaction with normalized marks.
    """
    for tag in tx.tags:
        if tag == mark_name or tag[0:len(mark_name+MARK_SEPERATOR)] == mark_name+MARK_SEPERATOR:
            tx = tx._replace(
                tags=tx.tags.difference([tag]),
                meta=metaset.add(tx.meta, mark_name, tag[len(mark_name+MARK_SEPERATOR):] or ''),
            )

    return tx


DEFAULT_APPLICABLE_ACCOUNT_TYPES = set(["Income", "Expenses", "Assets", "Liabilities", "Equity"])

def marked_postings(
        tx: Transaction,
        mark_name: str,
        applicable_account_types: Set[str] = DEFAULT_APPLICABLE_ACCOUNT_TYPES,
        allow_posting_level_mark: bool = True
    ):
    """
    Iterates over postings of the transaction, returning most specific mark value for applicable account types.

    Args:
        tx [Transaction]: transaction instance.
        mark_name [str]: the mark.
        applicable_account_types [Set[str]]: set of account types that must be considered, defaults to all five.
        allow_posting_level_mark [bool]: set to False if posting-level marks should raise error instead.
    Yields:
        list of mark values or None.
        posting.
        original posting.
        original transaction.
    """

    default_marks = metaset.get(tx.meta, mark_name)
    tx = tx._replace(
        meta=metaset.clear(tx.meta, mark_name)
    )

    for _posting in tx.postings:
        marks = metaset.get(_posting.meta, mark_name)
        posting = _posting._replace(
            meta=metaset.clear(_posting.meta, mark_name)
        )

        if(len(marks) > 0):
            yield marks, posting, _posting, tx
        elif(len(default_marks) > 0):
            if(posting.account.split(':')[0] not in applicable_account_types):
                yield None, posting, _posting, tx
            else:
                yield default_marks, posting, _posting, tx
        else:
            yield None, posting, _posting, tx

def sum_income(tx: Transaction) -> Inventory:
    total = Inventory()
    for posting in tx.postings:
        if(posting.account.split(':')[0] == "Income"):
            total.add_position(posting)
    return total

def sum_expenses(tx: Transaction) -> Inventory:
    total = Inventory()
    for posting in tx.postings:
        if(posting.account.split(':')[0] == "Expenses"):
            total.add_position(posting)
    return total




"""
Abstraction on top of beancount's meta `dict` to operate on set of values per single key.

Implements `set` operations, but immutable: `add`, `remove`, `discard`, `clear`

Implements new methods:
- `set` - overwrites with a new set.
- `get` - retrieves a set of values from meta.
- `reset` - tidy up suffixes.

Under the hood, each value is saved in a seperate key with unique suffix of digits. Order not guaranteed.
"""

from typing import List, Set, Union, Tuple
from copy import deepcopy

from beancount.core.data import Transaction, Posting, Meta
from beancount.core.inventory import Inventory

datatype_set = set
DIGITS_SET = datatype_set(['0','1','2','3','4','5','6','7','8','9'])

def contains_key(key: str, meta: str):
    """
    Determines whenever given meta is a mark, which may or may not be suffixed with numbers.

    Truthy meta examples for key "share":
    - share
    - share0
    - share900

    Falsey meta examples for key "share":
    - asdf
    - 42share
    - share42asdf

    Args:
        key [str]: the key.
        meta [str]: the meta.
    Return:
        Bool
    """
    if(key is meta):
        return True

    if(meta[0:len(key)] == key and datatype_set(meta[len(key):]) <= DIGITS_SET):
        return True

    return False

def get(meta: Meta, key: str) -> Set[str]:
    return [v for k,v in meta.items() if contains_key(key, k)]

def add(meta: Meta, key: str, value: str) -> Meta:
    copy = deepcopy(meta)
    safe_key: str

    if not (key in meta):
        safe_key = key
    else:
        suffix: int = 900 + len([k for k in meta if contains_key(key, k)])
        safe_key = key + str(suffix)

    copy[safe_key] = value
    return copy

def discard(meta: Meta, key: str) -> Meta:
    copy = deepcopy(meta)

    if (key in meta):
        del copy[key]

    return copy

def remove(meta: Meta, key: str) -> Meta:
    copy = deepcopy(meta)

    del copy[key]

    return copy

def clear(meta: Meta, key: str) -> Meta:
    copy = deepcopy(meta)

    for metakey in [k for k,_ in meta.items() if contains_key(key, k)]:
        del copy[metakey]

    return copy

def set(meta: Meta, key: str, new_set: Set[str]) -> Meta:
    copy = clear(meta, key)

    for elem in new_set:
        copy = add(meta, key, elem)

    return copy

def reset(meta: Meta, key: str) -> Meta:
    elements = get(meta, key)

    return set(meta, key, elements)



#!/usr/bin/env python3

from setuptools import setup

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='beancount_share',
    version='0.1.2',
    description='Plugin for Beancount to share expenses.',

    long_description=long_description,
    long_description_content_type='text/markdown',

    author='Akuukis',
    author_email='akuukis@kalvis.lv',
    download_url='https://pypi.python.org/pypi/beancount_share',
    license='GNU AGPLv3',
    package_data={'beancount_share': ['../README.md']},
    package_dir={'beancount_share': 'beancount_share'},
    packages=['beancount_share'],
    requires=['beancount (>2.0)'],
    url='https://github.com/Akuukis/beancount_share',
)
